/*
 MakeSparseFile v0.1
 Create a sparse file, update or copy a file to a sparse one.
 Copyright (C) 2008 Jrme Hod. 
	
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include <windows.h>
#include <tchar.h>
#include <stdio.h>

#if _UNICODE
#define _tcstoi _wtoi
#else
#define _tcstoi atoi
#endif

#define CHECK(isTrue) {if(!(isTrue)) { printLastErrorAndExit(TEXT("MakeSparseFile.c"), __LINE__);}}

static void printLastError(LPCTSTR context, int line) {
	LPVOID lpMsgBuf;

	FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |	FORMAT_MESSAGE_FROM_SYSTEM |	FORMAT_MESSAGE_IGNORE_INSERTS
		, NULL, GetLastError(), 0, (LPTSTR) &lpMsgBuf, 0, NULL);

	if(line) _tprintf( "%s [%d]: %s\n", context, line, (LPTSTR) lpMsgBuf);
	else  _tprintf( "%s: %s\n", context, (LPTSTR) lpMsgBuf);
	LocalFree(lpMsgBuf);
}
static void printLastErrorAndExit(LPCTSTR context, int line) {
	printLastError(context, line);
	exit(-1);
}


static void create(LPCTSTR filename,LPCTSTR sizespec) {
	LARGE_INTEGER position;
	position.QuadPart= _tcstoi(sizespec);
	switch(sizespec[_tcslen(sizespec)-1]) {
	case 'G': position.QuadPart<<= 10; 
	case 'M': position.QuadPart<<= 10;
	case 'K': position.QuadPart<<= 10; 
	}

	HANDLE fh = CreateFile(filename
		, GENERIC_READ | GENERIC_WRITE
		, 0
		, NULL
		, CREATE_NEW, FILE_ATTRIBUTE_NORMAL
		, NULL);
	if(fh==INVALID_HANDLE_VALUE) printLastErrorAndExit("CreateFile",0);

	DWORD bytesReturned;
	CHECK(DeviceIoControl(fh, FSCTL_SET_SPARSE, NULL, 0, NULL, 0, &bytesReturned, NULL));

	CHECK(SetFilePointerEx(fh, position, NULL, FILE_BEGIN));
	CHECK(SetEndOfFile(fh));

	CloseHandle(fh);
}

// NTFS compression unit size, not NTFS cluster size
//#define BLOCKSIZE 4096
#define BLOCKSIZE 65536

static void processAllocatedRanges(HANDLE srch, HANDLE dsth) {
	DWORD t0= GetTickCount();
	
	long* block= (long*)malloc(BLOCKSIZE); 
	CHECK(block);
	LONGLONG cnt= 0, zcnt= 0;

	DWORD bytesReturned;

	FILE_ALLOCATED_RANGE_BUFFER query;
	query.FileOffset.QuadPart = 0;
	CHECK(SetFilePointerEx(srch, query.FileOffset, &query.Length, FILE_END));
	if(srch!=dsth) {	
		CHECK(SetFilePointerEx(dsth, query.Length, NULL, FILE_BEGIN));
		CHECK(SetEndOfFile(dsth));
	}

	FILE_ALLOCATED_RANGE_BUFFER ranges[1024];
	for(;;) {
		BOOL completed= DeviceIoControl(srch, FSCTL_QUERY_ALLOCATED_RANGES, &query, sizeof(query), ranges, sizeof(ranges), &bytesReturned, NULL);
		CHECK(completed || GetLastError()==ERROR_MORE_DATA);

		int n= bytesReturned / sizeof(FILE_ALLOCATED_RANGE_BUFFER);
		//if(n) printf("%d ranges, from %I64d to %I64d\n",n,ranges[0].FileOffset.QuadPart,ranges[n-1].FileOffset.QuadPart+ranges[n-1].Length.QuadPart);
		LONGLONG rend;
		for (int i=0; i<n; ++i)  {
			FILE_ZERO_DATA_INFORMATION zr;
			zr.BeyondFinalZero.QuadPart= ranges[i].FileOffset.QuadPart;
			rend= zr.BeyondFinalZero.QuadPart+ ranges[i].Length.QuadPart;
			zr.FileOffset.QuadPart= -1;

			CHECK(SetFilePointerEx(srch, zr.BeyondFinalZero, NULL, FILE_BEGIN));
			if(dsth!=srch) CHECK(SetFilePointerEx(dsth, zr.BeyondFinalZero, NULL, FILE_BEGIN));

			printf("range %I64d - %I64d\r",zr.BeyondFinalZero.QuadPart,rend);
			for(; zr.BeyondFinalZero.QuadPart<rend; zr.BeyondFinalZero.QuadPart+= BLOCKSIZE) {
				CHECK(ReadFile(srch, block, BLOCKSIZE, &bytesReturned, NULL));
				if(dsth!=srch) CHECK(WriteFile(dsth, block, bytesReturned, &bytesReturned, NULL));

				for(bytesReturned/= 4;  (~--bytesReturned) && !block[bytesReturned];);
				if(~bytesReturned) {	
					if(zr.FileOffset.QuadPart>=0) {
						CHECK(DeviceIoControl(dsth, FSCTL_SET_ZERO_DATA, &zr, sizeof(zr), NULL, 0, &bytesReturned, NULL));
						zcnt+= zr.BeyondFinalZero.QuadPart-zr.FileOffset.QuadPart;
						zr.FileOffset.QuadPart= -1;
					}
				}
				else if(zr.FileOffset.QuadPart<0) zr.FileOffset.QuadPart= zr.BeyondFinalZero.QuadPart;

				//if((zr.BeyondFinalZero.QuadPart&((1<<20)-1))==0) printf("position: %I64d MB\r", zr.BeyondFinalZero.QuadPart>>20);
			}
			if(zr.FileOffset.QuadPart>=0) {
				zr.BeyondFinalZero.QuadPart= rend;
				CHECK(DeviceIoControl(dsth, FSCTL_SET_ZERO_DATA, &zr, sizeof(zr), NULL, 0, &bytesReturned, NULL));
				zcnt+= zr.BeyondFinalZero.QuadPart-zr.FileOffset.QuadPart;
			}

			cnt+= ranges[i].Length.QuadPart;
		}
		
		if(completed || !n) break;

		query.Length.QuadPart-= rend-query.FileOffset.QuadPart;
		query.FileOffset.QuadPart = rend;
	}

	free(block);
	printf("processed: %I64d bytes                    \ncleaned  : %I64d bytes\nduration: %d sec", cnt,zcnt,(GetTickCount()-t0)/1000);
}

static void update(LPCTSTR filename) {
	DWORD bytesReturned;

	HANDLE fh = CreateFile((LPTSTR) filename
		, GENERIC_READ | GENERIC_WRITE
		, 0
		, NULL
		, OPEN_EXISTING
		, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_NO_BUFFERING
		, NULL);
	if(fh==INVALID_HANDLE_VALUE) printLastErrorAndExit("CreateFile",0);

	CHECK(DeviceIoControl(fh, FSCTL_SET_SPARSE, NULL, 0, NULL, 0, &bytesReturned, NULL));

	processAllocatedRanges(fh,fh);

	CloseHandle(fh);
}

static void copy(LPCTSTR srcname, LPCSTR dstname) {
	DWORD bytesReturned;

	HANDLE srch = CreateFile((LPTSTR) srcname
		, GENERIC_READ
		, FILE_SHARE_READ
		, NULL
		, OPEN_EXISTING
		, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_NO_BUFFERING
		, NULL);
	if(srch==INVALID_HANDLE_VALUE) printLastErrorAndExit("CreateFile",0);

	HANDLE dsth = CreateFile((LPTSTR) dstname
		, GENERIC_READ | GENERIC_WRITE
		, 0
		, NULL
		, CREATE_NEW
		, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_NO_BUFFERING
		, NULL);
	if(dsth==INVALID_HANDLE_VALUE) printLastErrorAndExit("CreateFile",0);

	CHECK(DeviceIoControl(dsth, FSCTL_SET_SPARSE, NULL, 0, NULL, 0, &bytesReturned, NULL));

	processAllocatedRanges(srch,dsth);

	CloseHandle(dsth);
	CloseHandle(srch);
}

static void printUsage(LPCTSTR prgname) {
	_tprintf(TEXT("Usage: %s ...\n"
			"\t create <file> <size>{K|M|G}\n"
			"\t update <file>\n"
			"\t copy <source> <dest>\n"
			), prgname);
}

int _tmain(int argc, TCHAR *argv[]) {
	if(argc>=2) {
		if(_tcscmp(argv[1],TEXT("create"))==0) {
			if(argc==4) {
				create(argv[2],argv[3]);
				return 0;
			}
		}
		else if(_tcscmp(argv[1],TEXT("update"))==0) {
			if(argc==3) {
				update(argv[2]);
				return 0;
			}
		}
		else if(_tcscmp(argv[1],TEXT("copy"))==0) {
			if(argc==4) {
				copy(argv[2],argv[3]);
				return 0;
			}
		}
	}
	printUsage(argv[0]);
	return -1;
}
